// iodmgmt.c
// This file contains most of the code to manage and link iods to viod
//
// Changelog:
//
//  
// 2002xxxx - pl
//  - lots of clean ups
//
// Author: Adrien Lebre <adrien.lebre@imag.fr>
// Co-author: Pierre Lombard <pierre.lombard@imag.fr>
// Co-author:  Olivier Valentin <olivier.valentin@imag.fr>
// Date: 2002/07
// Licence: GPL

#include "config.h"
#include "default.h"

// defines structures and prototypes used
#include "iod_mgmt.h"

// defines msg_iod structure and ping request
#include "iodng.h"

// defines errors
#include "errno.h"
// defines  gethostbyname
#include <netdb.h>
extern int h_errno;

#include <sys/types.h>		// for socket()
#include <sys/socket.h>		// for socket(), inet_ntoa()
#include <arpa/inet.h>		// for inet_ntoa()

#include <sys/ioctl.h>		// for ioctl()
#include <net/if.h>		// for ioctl()

#include <unistd.h>		// for sleep(), getopt()
#include <getopt.h>		// for getopt()
#include <time.h>		// for nanosleep()

//#define SHOWINFO 1
//define VERBOSEPING 1  // Enables extra-verbose mode wrt ping

#include "ipstring.h"		// defines several macros like INFO, ERR ...
#include "logging.h"		// defines several macros like INFO, ERR ...

#include <ctype.h>		// for isblank()

#include <stdio.h>		// for sscanf()

#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#include <pthread.h>

#include "sure_sock.h"		// for sure_recv(), send_recv()

#include "nfspdctl.h"		// defines communication with nfspd

#define SOCK_INTERACT_ENV_VAR "HOME"
#define SOCK_INTERACT_NAME "iod_mgmt2iod_mgmt.ctl"

//#include <atomic.h>   // Required to update the index inside the Ping thread.

// Forward declarations
static int SetMode(int iMode);
static void inet_resolve(struct sockaddr_in *psin, const char *const host, const char *const port);
static void init_rand(void);
static int iods_AreYouAlive(unsigned char *eth0_mac, struct sockaddr_in iod, int iNbMessage, int wait);
static void SetConfigParameters(int iPause, int iNbMessage, int iWait);
static void GetConfigParameters(int *iPause, int *iNbMessage, int *iWait);
static char *xstrdup(const char *string);
static int IsInList(struct sockaddr_in TempIod, struct list_head *List);
static void Quit_IodMgmt(void);

/* **** *********************************** **** */

#define DIE(format,args...) do {		\
	fprintf(stderr, "Error: ");		\
	fprintf(stderr, format "\n",## args);	\
	exit(1);				\
} while (0)

/* **** *********************************** **** */

/* **** ****************************************************************************************************************** **** */
/* **** ********************************************** Global Variables ************************************************** **** */
/* **** ****************************************************************************************************************** **** */
static char *cPtrProgramName;	// The program name (should be iod_mgmt)

static int foreground = 0;	// Thr processus runs in foreground mode (or background by default).

static iod_info_t iods[MAX_IODS];	// The iods references
static struct list_head FreeIodList;	// Enables to manage the free space inside the iods tabs.
static struct list_head UsedIodList;	// Enables to parse each used space inside the iods tabs.

static pthread_mutex_t ConfigParameters_Mutex = PTHREAD_MUTEX_INITIALIZER;	// mutex on configuration parameters defined below
static int iByMutexPause = 300;	// Pause between two cycles of ping
static int iByMutexNbMessage = 3;	// Number of messages send in once to one iod.
static int iByMutexWait = 2;	// Pause beetween two ping to the same iod.

static int iNFSpMode = IOD_MGMT_MODE_DROP;

// not used but could be if I wanted to use pthread_cancel -- Adrien
// static pthread_t Ping_Thr;
// static pthread_t Interact_Thr

static int iods_nb = 0;		// number of iods
static int viods_nb = 0;	// number of viods
static int iods_udp_fd = -1;	// socket used to send UDP to iods (PING request)
static int sock_interact = -1;	// socket used to talk between Iod_mgmt instance and Interact Thread

/* **** ****************************************************************************************************************** **** */
/* **** ******************************************** Procedures and Functions ******************************************** **** */
/* **** ****************************************************************************************************************** **** */

		/* **** ********************************************************************************* **** */
		/* **** *************** Ping section (Cut and Paste from iod_ping.c) ******************** **** */
		/* **** ********************************************************************************* **** */

// Gets addresses (IP and MAC) for the specified interface
//
// Return value: 0 if success -1 in case of failure
// IN : itf             interface name (eth0, lo, ...)
// OUT: IPaddress       the IPv4 address (4 bytes)
// OUT: MACaddress      the HW address (6 bytes)
//
// TODO remove if is unused.
static int
get_if_addr(unsigned char *IPaddress, unsigned char *MACaddress, const char *const itf)
{
	struct ifreq ifr;
	int fd;

	Dprintf(D_IODMGR, "get_if_addr() start");

	fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (fd == -1) {
		printf("cannot open a socket errno=%d\n", errno);
		goto error;
	}

	strncpy(ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFADDR, &ifr)) {
		printf("cannot get the IP address of %s (SIOCGIFADDR)", ifr.ifr_name);
		goto error;
	}
	memcpy(IPaddress, &(((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr), 4);

	strncpy(ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFHWADDR, &ifr)) {
		printf("cannot get the HW address of %s (SIOCGIFHWADDR)", ifr.ifr_name);
		goto error;
	}
	memcpy(MACaddress, ifr.ifr_hwaddr.sa_data, 6);

	close(fd);

	return 0;

      error:
	memset(IPaddress, '\0', 4);
	memset(MACaddress, '\0', 6);
	return -1;
}

// Resolve the hostname 'host' and fills *psin
//
// OUT: psin            the socket structure to fill
// IN : host            the hostname
// IN : port            the port
//
// TODO Not used now - aimed at being used when I have implemented the
//   dynamic ADD iod command, for instance : iod_mgmt -a icluster25 (a
//   for add).
//
// FIXME: replace gethostbyname() by a MT safe gethostbyname_r()
static void
inet_resolve(struct sockaddr_in *psin, const char *const host, const char *const port)
{

	struct hostent *he;
	int numport;

	Dprintf(D_IODMGR, "init_resolve() function");

	memset(psin, '\0', sizeof (*psin));
	psin->sin_family = AF_INET;

	he = gethostbyname(host);
	if (he != NULL) {
		memcpy((char *) &(psin->sin_addr), he->h_addr, he->h_length);
	} else {
		DIE("cannot resolve host %s (errno=%d)", host, errno);
	}

	numport = atoi(port);
	if (0 < numport && numport < 65536) {
		psin->sin_port = htons(numport);
	} else {
		DIE("port must be between 1 and 65535 and is currently %d", numport);
	}
	psin->sin_port = htons(numport);
}

//
// Initialize the random numbers generator used to get a random seed
//
static void
init_rand(void)
{
	struct timeval tv;

	Dprintf(D_IODMGR, "init_rand() function");
	if (-1 == gettimeofday(&tv, NULL)) {
		DIE("cannot get time (errno=%d)", errno);
	}
	srand(tv.tv_usec);
}

//
// Ping one iod to check its status
//
// This function sends a message "are you alive" to one iod (specified by the sockaddr_in iod).
// "iNbMessage" request(s) are sent to the iod and separate by "wait" seconde(s) before to return its status.
//
// Return the status of the iod.
//
static int
iods_AreYouAlive(unsigned char *eth0_mac, struct sockaddr_in iod, int iNbMessage, int iWait)
{
	msg_iod_t ping_req, pong_rep;
	int bResult;
	struct sockaddr_in reply_sin;
	socklen_t reply_sin_len;

	Dprintf(D_IODMGR, "iods_AreYouAlive() function");
	init_rand();
	bResult = 0;
	// The socket descriptor, iods_udp_fd must be initialized by init_iods
	//
	if (iods_udp_fd < 0) {
		DIE("The UDP socket is not open - internal error");
	}

	Dprintf(D_IODMGR, "IOD_PING " NTODSTR " with %d bytes of data\n", NTOD(iod), MSG_IOD_PING_LEN);

	ping_req.pingreq.seq = rand() & 0xFFFF0000;
	while (0 != iNbMessage) {
		int sent, rcvd, ret;
		fd_set rfds;
		struct timeval tv;
		struct timeval tv_ping, tv_pong;
		int latency;

		if (iNbMessage > 0)
			iNbMessage--;

		// Sending PING REQUEST

		ping_req.magic = NFSP_MAGIC_MSG;
		ping_req.msgtype = NFSP_PING_REQ;
		ping_req.pingreq.seq++;

		ping_req.client_addr = 0;	// 0 => the iod will fill it
		ping_req.client_port = 0;	// 0 => the iod will fill it

		memcpy(ping_req.client_mac_addr, eth0_mac, 6);

		ping_req.server_addr = 0;	// fill by the kernel
		ping_req.server_port = 0;	// fill by the kernel

		// Send the pingreq through iods_udp_fd
		sent = sendto(iods_udp_fd, (char *) &ping_req, MSG_IOD_PING_LEN, MSG_NOSIGNAL, (struct sockaddr *) &iod, sizeof (struct sockaddr_in));
		if (sent == -1) {
			fprintf(stderr, "Error: cannot sendto (errno=%d)\n", errno);

			if (iNbMessage)
				sleep(iWait);
			continue;
		}
		// Waiting PONG_REPLY

		FD_ZERO(&rfds);
		FD_SET(iods_udp_fd, &rfds);
		tv.tv_sec = iWait;
		tv.tv_usec = 0;

		ret = gettimeofday(&tv_ping, NULL);
		if (ret)
			DIE("gettimeofday failed() (errno = %d)", errno);

		ret = select(iods_udp_fd + 1, &rfds, NULL, NULL, &tv);
		switch (ret) {
		case -1:
			DIE("cannot select() (errno=%d)", errno);
			break;

		case 0:
			// Timeout
#if VERBOSEPING
			Dprintf(D_IODMGR, "Reply for seq=0x%08lx: FAILED (timeout)\n", ping_req.pingreq.seq);
#endif				/* VERBOSEPING */
			break;

		default:
			reply_sin_len = sizeof (reply_sin);
			rcvd = recvfrom(iods_udp_fd, (char *) &pong_rep, MSG_IOD_PING_LEN, 0, (struct sockaddr *) &reply_sin, &reply_sin_len);

			ret = gettimeofday(&tv_pong, NULL);
			if (ret) {
				DIE("gettimeofday failed() (errno = %d)", errno);
			}
			latency = 1000000 * (tv_pong.tv_sec - tv_ping.tv_sec) + tv_pong.tv_usec - tv_ping.tv_usec;

			if (ping_req.pingreq.seq == pong_rep.pingreq.seq && pong_rep.magic == NFSP_MAGIC_MSG && pong_rep.msgtype == NFSP_PONG_REQ && rcvd == MSG_IOD_PING_LEN) {
#if VERBOSEPING
				Dprintf(D_IODMGR, "Reply from " NTODSTR ": seq=0x%08lx time=%d.%d ms\n", NTOD(reply_sin), ping_req.pingreq.seq, latency / 1000, latency % 1000);
#endif				/* VERBOSEPING */
				bResult++;
			} else {
#if VERBOSEPING
				Dprintf(D_IODMGR, "Reply from " NTODSTR " for seq=0x%08lx: FAILED"
				     " (seq=0x%08lx kind=0x%08x sent=%d rcvd=%d) time=%d.%d ms\n",
				     NTOD(reply_sin), ping_req.pingreq.seq, pong_rep.pingreq.seq, pong_rep.kind, sent, rcvd, latency / 1000, latency % 1000);
#endif				/* VERBOSEPING */
			}
			if (iNbMessage) {
				int to_sleep;	//in microseconds
				struct timespec req, rem;

				to_sleep = 1000000 * iWait - latency;

				memset(&req, 0, sizeof (req));
				req.tv_sec = to_sleep / 1000000;
				req.tv_nsec = (to_sleep % 1000000) * 1000;
				memset(&rem, 0, sizeof (rem));
				ret = nanosleep(&req, &rem);
				if (ret)
					DIE("nanosleep failed() (errno = %d)", errno);
			}
		}		//end of switch
	}			// end of while iNbMessage

	return (bResult > 0);	//return the status of the iod.
}

//
// Ping_Thr thread : Ping per cycle of "iPause" seconde(s) each iod contains in iod tabs
//
// This thread is launched by the Init_IodMgmt function during execution of the first instance of iodmgmt
//
static void *
Ping_Thr(void * dummy __attribute__((unused)))
{
	int i;
	int iPause, iWait, iNbMessage;
	unsigned char eth0_ip[4];
	unsigned char eth0_mac[6];

	Dprintf(D_IODMGR, "Ping_thr started");

	//
	// Retrieve IP and Mac source address
	if (-1 == get_if_addr(eth0_ip, eth0_mac, "eth0")) {
		DIE("cannot run get_if_addr()");
	} else {
		Dprintf(D_IODMGR, "Using interface eth0 (IP: %u.%u.%u.%u MAC:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n",
		     eth0_ip[0], eth0_ip[1], eth0_ip[2], eth0_ip[3], eth0_mac[0], eth0_mac[1], eth0_mac[2], eth0_mac[3], eth0_mac[4], eth0_mac[5]);
	}

	//
	// Retrieve the configuration parameters
	GetConfigParameters(&iPause, &iNbMessage, &iWait);

	// Ping All iods
	while (1) {
		// ping all iods
		for (i = 0; i < iods_nb; i++) {
			GetConfigParameters(&iPause, &iNbMessage, &iWait);
			if (iods_AreYouAlive(eth0_mac, iods[i].iod, iNbMessage, iWait)) {
				Dprintf(D_IODMGR, "iod " NTODSTR " is alive", NTOD(iods[i].iod));
				iods[i].check_ok = 1;
			} else {
				Dprintf(D_IODMGR, "iod " NTODSTR " is down", NTOD(iods[i].iod));
				iods[i].check_ok = 0;

				// TODO : unbind / free ?
			}
			if (-1 == gettimeofday(&iods[i].check_time, NULL)) {
				Dprintf(L_WARNING, "Ping_Thr()::gettimeofday() failed");
			}
		}

		Dprintf(D_IODMGR, "Ping_Thr is in pause for %d seconds", iPause);
		sleep(iPause);

		// This code replace sleep(iPause)
		// The interest is to decrease the granularity of the deplay between each cycle.
		// By the way, if the pause parameter has been modified the new value is assigned.
		// and the delay is lower.
		// TODO : Clean this nasty things :(, maybe use sigalarm :
		// something like
		// int  sigalarm(funct) where function is equal to do a cycle and rearm the alarm.
		/*      do
		   {
		   atomic_t i;
		   TODO Atomic test between i and iPause
		   //for(i=atomic_set(i,0);i>=iPause;atomic_inc(&i))
		   {
		   GetConfigParameters(&iPause,&iNbMessage,&iWait);
		   sleep(1);
		   }
		   }
		   while(0); */
		//sleep(iPause);

	}			// End of While(1)
}

		/* **** ********************************************************************************* **** */
		/* **** ***************************** Interact Functions ******************************** **** */
		/* **** ********************************************************************************* **** */
//
// Set the value of different ping parameters
// iNbMessage equal to the number of ping for each "life test".
// iPause the delay before a new cycle (a cycle consist to iNbMessage ping(s) on each iod)
// iWait the delay between two simple ping.
//
//  A mutex appear in order to be consitent between Interact_Thr and Ping_Thr thread
static void
SetConfigParameters(int iNbMessage, int iPause, int iWait)
{
	pthread_mutex_lock(&ConfigParameters_Mutex);

	iByMutexPause = iPause;	// Pause between two cycles of ping
	iByMutexNbMessage = iNbMessage;	// Number of messages send in once to one iod.
	iByMutexWait = iWait;	// Pause beetween two ping to the same iod.

	pthread_mutex_unlock(&ConfigParameters_Mutex);
}

//
// Equivalent to SetConfigParameters but take a char* argument which is going to be decompose in three values.
// Return 0 if it's ok.
//
static int
SetConfigParametersByLine(char *cPtrLine)
{
//      char cTabBuf[MAXLINELEN + 1];
	int iNbMessage, iPause, iWait;
	if (sscanf(cPtrLine, "%d %d %d\n", &iNbMessage, &iWait, &iPause) != 3) {
		Dprintf(D_IODMGR, "The format of the string (\"%s\") is invalid\n", cPtrLine);
		return -1;
	}
	SetConfigParameters(iNbMessage, iPause, iWait);
	return 0;
}

//
// Equivalent to SetConfigParameters (just above) but msut be call by the second (or third or ...) instance of iodmgmt.
// This function tries to discuss with Interact_Thr thread in order to set the Ping_Thr's conifguration parameters.
//
// Sock_fd is the socket use to speak with the remote thread (Interact_Thr).
// cPtrline contains the new values to assign..
//
// Notes:
//
int
RemoteSetConfigParameters(int Sock_fd, char *cPtrLine)
{
	int iResult;
	interact_request_t Request;

	Request.iType = IOD_MGMT_CMD_SET_PING_PARAMETERS;
	strcpy(Request.cTabParam, cPtrLine);

	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteSetConfigParameters()::sure_write() failed errno=%d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (int), 0) != sizeof (int)) {
		Dprintf(L_WARNING, "RemoteSetConfigParameters()::sure_recv() failed errno=%d", errno);
		return -1;
	}

	if (iResult >= 0) {
		fprintf(stdout, "The new configuration parameters have been set.\n");
#if 0
		printf(stdout, "The configuration parameters are :\n"
		       "  loop delay: %d seconds\n" "  send %d pings in a row\n" "  wait %d seconds before sending 2 pings\n", iPause, iNbMessage, iWait);
#endif				/* 0 */

	} else {
		fprintf(stdout, "Error: cannot assign the new configuration parameters.\n");
	}

	return iResult;
}

/*azerty J en suis ici !! */

static void
GetConfigParameters(int *iPause, int *iNbMessage, int *iWait)
{
	pthread_mutex_lock(&ConfigParameters_Mutex);
	*iPause = iByMutexPause;	// Pause between two cycles of ping
	*iNbMessage = iByMutexNbMessage;	// Number of messages send in once to one iod.
	*iWait = iByMutexWait;	// Pause beetween two ping to the same iod.
	pthread_mutex_unlock(&ConfigParameters_Mutex);
}

static int
SendConfigParameters(int Sock_fd)
{
	int iResult = -1;
	int iPause, iNbMessage, iWait;

	GetConfigParameters(&iPause, &iNbMessage, &iWait);

	if (sure_write(Sock_fd, &iPause, sizeof (iPause)) != sizeof (iPause)) {
		Dprintf(L_WARNING, "SendConfigParameters()::sure_write() failed errno = %d", errno);
		return -1;
	}
	if (sure_write(Sock_fd, &iNbMessage, sizeof (iNbMessage)) != sizeof (iNbMessage)) {
		Dprintf(L_WARNING, "SendConfigParameters()::sure_write() failed errno = %d", errno);
		return -1;
	}
	if (sure_write(Sock_fd, &iWait, sizeof (iWait)) != sizeof (iWait)) {
		Dprintf(L_WARNING, "SendConfigParameters()::sure_write() failed errno = %d", errno);
		return -1;
	}

	return iResult;
}

int
RemoteDisplayConfigParameters(int Sock_fd)
{
	int iResult;
	int iPause, iNbMessage, iWait;
	interact_request_t Request;

	Request.iType = IOD_MGMT_CMD_GET_PING_PARAMETERS;
	Request.cTabParam[0] = '\0';
	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteDisplayConfigParameters()::sure_write() failed errno=%d", errno);
		return -1;
	}
	// Retrieve each parameter.
	if (sure_recv(Sock_fd, &iPause, sizeof (int), MSG_NOSIGNAL) != sizeof (int)) {
		Dprintf(L_WARNING, "RemoteDisplayConfigParameters()::sure_recv() failed errno=%d", errno);
		return -1;
	}
	if (sure_recv(Sock_fd, &iNbMessage, sizeof (int), MSG_NOSIGNAL) != sizeof (int)) {
		Dprintf(L_WARNING, "RemoteDisplayConfigParameters()::sure_recv() failed errno=%d", errno);
		return -1;
	}
	if (sure_recv(Sock_fd, &iWait, sizeof (int), MSG_NOSIGNAL) != sizeof (int)) {
		Dprintf(L_WARNING, "RemoteDisplayConfigParameters()::sure_recv() failed errno=%d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (int), MSG_NOSIGNAL) != sizeof (int)) {
		Dprintf(L_WARNING, "RemoteDisplayConfigParameters()::sure_recv() failed errno=%d", errno);
		return -1;
	}

	if (iResult >= 0)
		// Display config parameters
		fprintf(stdout, "The configuration parameters are :\n"
			"The cycle restart all %d seconds\n" "Each time, %d ping are sent with %d seconds between each request.\n", iPause, iNbMessage, iWait);
	else
		fprintf(stdout, "Error : cannot retrieve the configuration parameters.\n");

	return iResult;
}

//
// Set the number of viod in nfsp kernel.
// A copy of this value is stored in viods_nb variable during all execution of the first instance of iodmgmt.
//
// iNbViod contains the new value to set...
// Return the result of this function -1 if something was wrong in nfspd_setnbviods or the new viods's number.
//
// Notes:
//      You have to be in DROP MODE in order to call this function
//      The next function "SetNbViodByLine" is similar but take a char* argument.
int
SetNbViod(int iNbViod)
{
	//TODO : Log this operation in a specific file

	Dprintf(D_IODMGR, "SetNbViod() function");
#if IOD_MGMT_WITH_NFSP
	if (nfspd_setnbviods(iNbViod) < 0) {
		Dprintf(L_WARNING, "nfspd_setnbviods() failed");
		return -1;
	}
#endif
	viods_nb = iNbViod;
	return viods_nb;
}

int
SetNbViodByLine(char *cPtrLine)
{

	Dprintf(D_IODMGR, "Try to SetNbViodByLine with %s value in char format and %d in int format\n", cPtrLine, atoi(cPtrLine));
	return SetNbViod(atoi(cPtrLine));	//FIXME : How could we be sure about the atoi behaviour !!!!

}

//
// Set the number of viod in nfsp kernel but msut be call by the second (or third or ...) instance of iodmgmt.
// This function tries to discuss with Interact_Thr thread in order to set the viod's number.
//
// Sock_fd is the socket use to speak with the remote thread (Interact_Thr).
// cPtrline contains the new value to assign..
//
// Notes:       You have to be in DROP MODE in order to call this function
//
int
RemoteSetNbViod(int Sock_fd, char *cPtrLine)
{
	interact_request_t Request;
	int iResult;

	Dprintf(D_IODMGR, "You try to set the viod's number with %s value\n", cPtrLine);
	Request.iType = IOD_MGMT_CMD_SET_VIODS_NB;
	strcpy(Request.cTabParam, cPtrLine);

	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteSetNbViod()::sure_write() failed errno=%d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteAddIod()::sure_recv() failed errno=%d", errno);
		return -1;
	}

	if (iResult >= 0)
		fprintf(stdout, "NFSp kernel is set to %s viod(s).\n", Request.cTabParam);
	else
		fprintf(stdout, "There was something wrong during the viod set operation.\n" "The string (%s) contains some mistakes or something is wrong with NFSp.\n", Request.cTabParam);
	return iResult;
}

//
// Return the viods_nb
//
// Notes: Seems not used !!!
//
int
GetNbViods(void)
{
	Dprintf(D_IODMGR, "GetNbViod() function");
	return viods_nb;
}

//
// Send the iods tab (managed by Interact_Thr and Ping_Thr) to one other iodmgmt instance.
//
// Sock_fd is the socket use to speak with the other instance of iodmgmt.
// List is the linked list (almost UsedIodList).
//
// Notes:       This function is called by the first iodmgmt's (by Interact_Thr thread).
//
int
SendList(int Sock_fd, struct list_head *List)
{
	struct list_head *PtrIndex, *PtrTemp;	// Used to parse the list
	iod_info_t *PtrIodTemp;	// Used to retrieve a temporary copy of an element of the iods array

	Dprintf(D_IODMGR, "SendList() function");

	// Send the number of iods
	if (sure_write(Sock_fd, &iods_nb, sizeof (iods_nb)) != sizeof (iods_nb)) {
		Dprintf(L_WARNING, "SendList()::sure_write() failed errno=%d", errno);
		return -1;
	}
	// Send the number of iods
	if (sure_write(Sock_fd, &viods_nb, sizeof (viods_nb)) != sizeof (viods_nb)) {
		Dprintf(L_WARNING, "SendList()::sure_write() failed errno=%d", errno);
		return -1;
	}
	// Parse the Linked List.
	list_for_each_safe(PtrIndex, PtrTemp, List) {
		// Retrieve the next iod space and send it on socket
		PtrIodTemp = list_entry(PtrIndex, iod_info_t, FromAndToList);
		if (sure_write(Sock_fd, PtrIodTemp, sizeof (*PtrIodTemp)) != sizeof (*PtrIodTemp)) {
			Dprintf(L_WARNING, "SendList()::sure_write() failed errno=%d", errno);
			return -1;
		}
	}
	return 1;
}

#if 0
// List all Informations about iods contains in the list "List".
// List : iods_info iods tabs, iods_nb and number of viods.
//
// List is the Linked List to display (almost UsedIodList).
//
//Note:     This function must be call by the first iodmgmt instance.
//              Seems not used !!!!
//
static void
DisplayIodsList(struct list_head *List)
{
	struct list_head *PtrIndex, *PtrTemp;
	iod_info_t *PtrIodTemp;

	Dprintf(D_IODMGR, "DisplayIodsList() function");

	fprintf(stdout, "There are %d iods for %d viods.\n", iods_nb, viods_nb);

	list_for_each_safe(PtrIndex, PtrTemp, List) {
		PtrIodTemp = list_entry(PtrIndex, iod_info_t, FromAndToList);
		if (PtrIodTemp->check_ok) {
			fprintf(stdout, "iod " NTODSTR " is alive and belongs to viod %d\n", NTOD(PtrIodTemp->iod), PtrIodTemp->iRelated_viod);
		} else {
			fprintf(stdout, "iod " NTODSTR "is down and belongs to viod %d\n", NTOD(PtrIodTemp->iod), PtrIodTemp->iRelated_viod);
		}
	}
}
#endif

//
// Equivalent to DisplayIodsList (just above) but have to be used by an second instance of iodmgmt.
// List all Informations about iods contains in the list "List".
// List : iods_info iods tabs, iods_nb and number of viods.
//
// Sock_fd is the socket use to speak with the remote thread (Interact_Thr).
// List is the Linked List to display (almost UsedIodList).
//
// Notice: This function must be called by another iodmgmt instance from the first.
//
int
RemoteDisplayIodsList(int Sock_fd)
{
	int iResult;

	Dprintf(D_IODMGR, "RemoteDisplayIodsList() function");
	// Send the request to Interact_Thr (to the first iodmgmt instance...)
	do {
		interact_request_t Request;

		Request.iType = IOD_MGMT_CMD_LIST_IOD;
		Request.cTabParam[0] = '\0';
		if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
			Dprintf(L_WARNING, "DisplayIodsRemoteList()::sure_write() failed errno = %d\n", errno);
			return -1;
		}
	} while (0);

	// Retrieve and display all informations about the iods and viods
	//  TODO : implement a clean RemoteGetNbViods function.
	do {
		iod_info_t IodTemp;
		int i;
		int iTemp_iods_nb = 0;	// The length of the list you're going to receive.
		int iTemp_viods_nb = 0;	// The number of the viod.

		if (sure_recv(Sock_fd, &iTemp_iods_nb, sizeof (iTemp_iods_nb), MSG_NOSIGNAL) != sizeof (iTemp_iods_nb)) {
			Dprintf(L_WARNING, "RemoteDisplayIodsList()::sure_recv() failed errno = %d", errno);
			return -1;
		}

		if (sure_recv(Sock_fd, &iTemp_viods_nb, sizeof (iTemp_viods_nb), MSG_NOSIGNAL) != sizeof (iTemp_viods_nb)) {
			Dprintf(L_WARNING, "RemoteDisplayIodsList()::sure_recv() failed errno = %d", errno);
			return -1;
		}
		fprintf(stdout, "There are %d iods for %d viods.\n", iTemp_iods_nb, iTemp_viods_nb);

		for (i = 0; i < iTemp_iods_nb; i++) {

			// FIXME: check some problems with sure_recv()
			if (sure_recv(Sock_fd, &IodTemp, sizeof (IodTemp), MSG_NOSIGNAL) != sizeof (IodTemp)) {
				Dprintf(L_WARNING, "RemoteDisplayIodsList()::sure_recv() failed errno=%d", errno);
				return -1;
			}

			if (IodTemp.check_ok) {
				fprintf(stdout, "iod " NTODSTR " is alive and belongs to viod %d\n", NTOD(IodTemp.iod), IodTemp.iRelated_viod);
			} else {
				fprintf(stdout, "iod " NTODSTR " is down and belongs to viod %d\n", NTOD(IodTemp.iod), IodTemp.iRelated_viod);
			}
		}
	} while (0);

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteDisplayIodsList()::sure_recv() failed errno = %d\n", errno);
		return -1;
	}
	return iResult;
}

//
// Add a new iod inside the iods buffer
// Return the number of iods
// or -1 if there is something wrong.
//
// cPtrLine contains the string parameter ("host:port viod")
//
// Note:
//      This function must be call by Interact_Thr (first iodmgmt instance).
//
int
AddIod(const char *cPtrLine)
{

	char cTabBuf[MAXLINELEN + 1];
	char *host = NULL;
	char *port = NULL;
	int iTempRelatedViod;

	Dprintf(D_IODMGR, "AddIod() function");

	// Try to retieve the different parameters
	if (sscanf(cPtrLine, "%s %d\n", cTabBuf, &iTempRelatedViod) != 2) {
		Dprintf(L_WARNING, "The format of the parameter (%s) is invalid (must be host:port)", cPtrLine);
		return -1;
	}
	// Check if the viod is coherent.
	if (iTempRelatedViod >= viods_nb) {
		Dprintf(L_WARNING, "You are trying and binding an iod to a nonexistant viod");
		return -1;
	}
	// cTabBuf contain "host:port" so it is now splitted into host & port on the ':'
	port = strchr(cTabBuf, ':');
	if (NULL == port) {
		Dprintf(L_WARNING, "The format of the parameter (%s) is invalid (must be host:port)", cPtrLine);
		return -1;
	}
	host = cTabBuf;
	*port = '\0';
	port++;

	Dprintf(D_IODMGR, "iod %d %s : %s", iods_nb, host, port);
	do {
		iod_info_t *PtrIodTmp;
		struct sockaddr_in IodTmp;

		// fill the sockaddr struct
		inet_resolve(&IodTmp, host, port);

		// If the iod is already in UsedIodList. We couldn't "re"add.
		if (!IsInList(IodTmp, &UsedIodList)) {
			// Have to search the next free iod
			if (FreeIodList.next == &FreeIodList) {
				Dprintf(L_WARNING, "Iods tab is full\n");	// correspond to iods_nb=MAX_IODS
				return -1;
			}
			// Retrieve the next free space
			PtrIodTmp = list_entry(FreeIodList.next, iod_info_t, FromAndToList);

			// fill fields
			memcpy(&PtrIodTmp->iod, &IodTmp, sizeof (IodTmp));
			PtrIodTmp->check_ok = 0;
			memset(&PtrIodTmp->check_time, '\0', sizeof (PtrIodTmp->check_time));
			PtrIodTmp->iRelated_viod = iTempRelatedViod;

			//TODO : Log this operation in a specific file
#if IOD_MGMT_WITH_NFSP
			// Add iod in nfsp kernel
			PtrIodTmp->iKernelHandle = nfspd_addiod(&PtrIodTmp->iod);
			if (PtrIodTmp->iKernelHandle < 0) {
				Dprintf(L_WARNING, "something went wrong in nfspd_addiod() function,(errno = %d).\n", errno);
				return -1;
			}
			// Bind iod in nfsp kernel
			if (0 > nfspd_bindiod(PtrIodTmp->iRelated_viod, PtrIodTmp->iKernelHandle)) {
				Dprintf(L_WARNING, "something went wrong in nfspd_bindiod,(errno = %d).\n", errno);
				if (0 > nfspd_freeiod(PtrIodTmp->iKernelHandle)) {
					DIE("Something is wrong with kernel.\n" "NFSp is in an inconsisten state !!!\n" "Please restart iodmgmt and flush the iods configuration.");
					// FIXME: it'd be better panic() now !
					// We have to let the file in the used list as we cannot free it
				}
				return -1;
			}
#endif
			//Extract from Free list
			list_del(&(PtrIodTmp->FromAndToList));
			// Add in Used List
			list_add_tail(&(PtrIodTmp->FromAndToList), &UsedIodList);
			return ++iods_nb;
		} else {
			Dprintf(L_WARNING, "The iod %s:%s is already present.\n", host, port);
		}
	}
	while (0);
	return -1;
}

//
// Equivalent to AddIod (just above) but have to be used by an second instance of iod_mgmt.
// Add a new iod inside the iods buffer
// Return the number of iods or -1 if there is something wrong.
//
// Sock_fd is the socket use to speak with the remote thread (Interact_Thr).
// cPtrline contains the new value to assign..
//
//Note: Must be call by another instance of iodmgmt
//
int
RemoteAddIod(int Sock_fd, const char *cPtrLine)
{
	interact_request_t Request;
	int iResult;

	Dprintf(D_IODMGR, "RemoteAddIod() function");

	Request.iType = IOD_MGMT_CMD_ADD_IOD;
	strcpy(Request.cTabParam, cPtrLine);

	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteAddIod()::sure_write() failed errno = %d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteAddIod()::sure_write() failed errno = %d", errno);
		return -1;
	}

	if (iResult >= 0) {
		fprintf(stdout, "The iod (%s) has been added and linked" "to the corresponding viod.\n", Request.cTabParam);
	} else {
		fprintf(stdout, "Something wrong happened during the addition operation.\n" "The parameter (%s) may contain some mistakes or the iods table may be full.\n", Request.cTabParam);
	}

	return iResult;
}

//
// Add new iod and bind it from a file
// This function parses the configuration file.
//
// Sock_fd is the socket use to speak with the remote thread (Interact_Thr).
// cPtrFile is the configuration file to parsed
//
// Notes:       order matters, no blank line, no comment line
//              A configuration file matches the following pattern : host:port viod [#comment]\n
//              XXX this primitive parser should be cleanly rewritten
//
static int
RemoteAddIodsFromFile(int Sock_fd, const char *cPtrFile)
{
	int iResult = 0;
	FILE *f;
	char buf[MAXLINELEN + 1];
	int line_no = 0;

	Dprintf(D_IODMGR, "RemoteAddIodsFromFile() function");

	Dprintf(D_IODMGR, "parsing the config file: %s", cPtrFile);
	f = fopen(cPtrFile, "r");
	if (NULL == f) {
		Dprintf(L_WARNING, "cannot open %s errno=%d", cPtrFile, errno);
		return -1;
	}

	while (NULL != fgets(buf, MAXLINELEN, f)) {
		if (iResult >= MAX_IODS) {
			Dprintf(L_WARNING, "too many iod servers in %s (MAX_IODS=%d)", cPtrFile, MAX_IODS);
			return -1;
		}
		// If a # is found the line is over
		// and skip the blank lines
		if ((!strchr(buf, '#')) && (strlen(buf) > 0)) {
			if ((iResult = RemoteAddIod(Sock_fd, buf)) < 0) {
				Dprintf(L_WARNING, "line %d of %s: MUST be host:port RelatedVIOD.\nPlease check your file", line_no, cPtrFile);
				return -1;
			}
		} else {
			Dprintf(D_IODMGR, "blank or comment line found: skipping it");
		}
	}			// process next line

	if (EOF == fclose(f)) {
		Dprintf(L_WARNING, "fclose() failed... errno=%d", errno);
	}
	return iResult;
}

//
// Test if a specific iod is already present in a list (specified by List).
//
// Return an boolean/integer (0 or 1)
//
//Note: Must be call by the first iodmgmt instance (Interact_Thr thread).
//
int
IsInList(struct sockaddr_in TempIod, struct list_head *List)
{
	int found = 0;
	iod_info_t *PtrIodTemp = NULL;
	struct list_head *PtrIndex, *PtrTemp;

	list_for_each_safe(PtrIndex, PtrTemp, List) {
		PtrIodTemp = list_entry(PtrIndex, iod_info_t, FromAndToList);	// Retrieve the next element.
		if (memcmp(&(PtrIodTemp->iod), &TempIod, sizeof (struct sockaddr_in)) == 0) {
			Dprintf(D_IODMGR, "The iod " NTODSTR " has been found", NTOD(TempIod));
			found = 1;
			break;
		}
	}
	return found;
}

//
// Free one iod inside the iods buffer.
// Unbind if it is.
// Return the number of iods
// or -1 if there is something wrong (If the iod doesn't exist or isn't appearing in the iod list).
//
// Notice:  If the iod to free is the last one of a viod list, then NFSP is switched to DROP MODE.
//
int
FreeIod(const char *cPtrLine)
{
	char cTabBuf[MAXLINELEN + 1];
	char *host = NULL;
	char *port = NULL;
	iod_info_t IodToDelete;	// Iod fill by inet_resolve from host:port
	iod_info_t *PtrIodTemp = NULL;	// IodTemp to parse the iod list
	int found;
	struct list_head *PtrIndex, *PtrTemp;
	int ret;

	sscanf(cPtrLine, "%s\n", cTabBuf);

	// buf now hopefully contain "host:port" so it is now splitted into
	// host & port on the ':'
	port = strchr(cTabBuf, ':');
	if (NULL == port) {
		return -1;	// Error, the line must be host:port
	}
	host = cTabBuf;
	*port = '\0';
	port++;

	// Retrieve the sockaddr_in structure to compare with the iods contain in the list "List".
	inet_resolve(&(IodToDelete.iod), host, port);

	// Have to search the iod corresponding to IodTmp
	found = 0;
	list_for_each_safe(PtrIndex, PtrTemp, &UsedIodList) {
		PtrIodTemp = list_entry(PtrIndex, iod_info_t, FromAndToList);
		if (!memcmp(&(PtrIodTemp->iod), &(IodToDelete.iod), sizeof (IodToDelete.iod))) {
			Dprintf(D_IODMGR, "The iod %s:%s has been found", host, port);
			found = 1;
			break;
		}
	}

	if (!found) {
		Dprintf(D_IODMGR, "The iod %s:%s was not found.s", host, port);
		return -1;
	}
#if IOD_MGMT_WITH_NFSP
	if (0 > nfspd_unbindiod(PtrIodTemp->iKernelHandle)) {
		Dprintf(L_WARNING, "FreeIod()::nfspd_unbindiod() errno = %d\n", errno);
		return -1;
	}
	if (0 > nfspd_freeiod(PtrIodTemp->iKernelHandle)) {
		Dprintf(L_WARNING, "FreeIod()::nfspd_freeiod() iod is nfsp server but not bound... errno = %d\n" "                           Trying to fix things...\n", errno);

		// maybe we have to rebind to the last viod... (consistency state!!!)
		if (0 > nfspd_bindiod(PtrIodTemp->iRelated_viod, PtrIodTemp->iKernelHandle))
			DIE("FreeIod()::nfspd_bindiod() NFSP in inconsistent state so I die... errno = %d\n", errno);
		// FIXME: it'd be better panic() now !
		// We have to let the file in the used list as we cannot free it
		return -1;
	}

	ret = nfspd_get_viod_size(PtrIodTemp->iRelated_viod);
	if (ret < 0) {
		Dprintf(L_WARNING, "FreeIod()::nfspd_get_viod_size() errno=%d\n", errno);
		return -1;
	} else if (ret == 0) {
		//switch in DROP MODE
		if (SetMode(IOD_MGMT_MODE_DROP)) {
			Dprintf(L_WARNING, "FreeIod()::SetMode() cannot go into DROP MODE\n");
		}
	}
#endif

	// Everythings seems ok... so now, we can remove this iod
	list_del(&(PtrIodTemp->FromAndToList));	//Extract from UsedIodList
	list_add_tail(&(PtrIodTemp->FromAndToList), &FreeIodList);	// Add in Free List
	iods_nb--;
	return iods_nb;
}

//
// Equivalent to FreeIod (just above) but have to be used by an second instance of iod_mgmt.
// Return the number of iods
// or -1 if there is something wrong (If the iod doesn't exist or isn't appearing in the iod list).
//
// Sock_fd is the socket use to speak with the remote thread (Interact_Thr).
// cPtrLine is the iod ("host:port" format) to remove.
//
int
RemoteFreeIod(int Sock_fd, const char *cPtrLine)
{
	int iResult;
	interact_request_t Request;

	Request.iType = IOD_MGMT_CMD_FREE_IOD;
	strcpy(Request.cTabParam, cPtrLine);
	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteFreeIod()::sure_write() failed errno = %d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteFreeIod()::sure_recv() failed errno = %d", errno);
		return -1;
	}

	if (iResult >= 0)
		fprintf(stdout, "The iod (%s) has been deleted.\n", Request.cTabParam);
	else
		fprintf(stdout, "Error : The iod (%s) hasn't been found" "and couldn't be deleted.\n", Request.cTabParam);

	return iResult;
}

//
// Switch in DROP mode, delete all iods and set viods_nb to 0..
// Must be call by Interact_Thr()
// Return the number's iod (must be equal to zero!!!).
int
Flush(void)
{
	struct list_head *PtrIndex, *PtrTemp;
	iod_info_t *PtrIodTemp;	// enables to retrieve the viod which must be unbind.

	if (nfspd_flush() < 0) {
		Dprintf(L_WARNING, "something nasty happened in nfspd_flush in function Flush,(errno = %d).\n", errno);
		return -1;
	}

	if (SetMode(IOD_MGMT_MODE_DROP) < 0) {
		Dprintf(L_WARNING, "something nasty happened in SetMode(DROP) in function Flush,(errno = %d).\n", errno);
		return -1;
	}

	list_for_each_safe(PtrIndex, PtrTemp, &UsedIodList) {
		PtrIodTemp = list_entry(PtrIndex, iod_info_t, FromAndToList);	// Retrieve the next element.
		list_del(&(PtrIodTemp->FromAndToList));	//Extract from Free list
		list_add_tail(&(PtrIodTemp->FromAndToList), &FreeIodList);	// Add in Free List
		iods_nb--;
	}
	return iods_nb;
}

//
// Equivalent to Flush (just above) but have to be used by an second instance of iod_mgmt.
// Switch in DROP mode, delete all iods and set viods_nb to 0.
// Sock_fd is th socket use to speak with Interact_Thr.
//
// Return the number of iod (must be 0!!!) else something was wrong.
//
int
RemoteFlush(int Sock_fd)
{
	int iResult;
	interact_request_t Request;

	Request.iType = IOD_MGMT_CMD_FLUSH_IOD;
	Request.cTabParam[0] = '\0';
	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteFlush()::sure_write() failed errno=%d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteFlush()::sure_recv() failed errno=%d", errno);
		return -1;
	}

	if (iResult == 0)
		fprintf(stdout, "NFSp is in DROP mode (no iods)\n");
	else
		fprintf(stdout, "Error : cannot flush nfsp.\n");

	return iResult;
}

//
// Change NFSp Mode.
//  IOD_MGMT_MODE_FORWARD or IOD_MGMT_MODE_DROP
//
// Note: must be called by Interact_Thr (first iodmgmt instance).
//
int
SetMode(int iMode)
{
	int iResult = -1;

	switch (iMode) {
	case IOD_MGMT_MODE_DROP:
#if IOD_MGMT_WITH_NFSP
		iResult = nfspd_drop();	//TODO DIE if drop doesn't run...
#endif
		iNFSpMode = iMode;
		break;
	case IOD_MGMT_MODE_FORWARD:
		if (viods_nb < 0) {
			return -1;
		}
#if IOD_MGMT_WITH_NFSP
		iResult = nfspd_forward();
#endif
		iNFSpMode = iMode;
		break;
	default:
		Dprintf(L_WARNING, "The mode %d is not defined", iMode);
	}
#if IOD_MGMT_WITH_NFSP
	if (iResult < 0) {
		Dprintf(L_WARNING, "cannot change NFSP mode errno = %d\n", errno);
	}
#endif
	return iResult;
}

//
// Equivalent to SetMode (just above) but have to be used by an second instance of iod_mgmt.
// Switch  mode (between IOD_MGMT_MODE_FORWARD and IOD_MGMT_MODE_DROP).
//
// Sock_fd is the socket use to speak with Interact_Thr.
// iMode is the new mode to assign (IOD_MGMT_MODE_FORWARD = 1 and IOD_MGMT_MODE_DROP = 0).
//
// Return 0 if the function succeded (<0 in the other case).
//
int
RemoteSetMode(int Sock_fd, int iMode)
{
	interact_request_t Request;
	int iResult;

	Request.iType = IOD_MGMT_CMD_SET_NFSP_MODE;
	Request.cTabParam[0] = (char) iMode;	//TODO : change this, isn't specially clean ;) but...

	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteSetMode()::sure_write() failed errno = %d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteSetMode()::sure_recv() failed errno = %d", errno);
		return -1;
	}

	if (iResult >= 0) {
		if (iMode == IOD_MGMT_MODE_DROP) {
			fprintf(stdout, "NFSp server is set to DROP mode\n");
			return iResult;
		} else if (iMode == IOD_MGMT_MODE_FORWARD) {
			fprintf(stdout, "NFSp server is set to FORWARD mode\n");
			return iResult;
		}
	}

	fprintf(stdout, "There was something wrong during the mode set operation.\n" "The mode (%s) is unknown or something is wrong with NFSp.\n", Request.cTabParam);
	return iResult;
}

//
// GetMode, have to be used by an second instance of iod_mgmt.
// Return the currently  mode (either IOD_MGMT_MODE_FORWARD or IOD_MGMT_MODE_DROP).
//
// Sock_fd is the socket use to speak with Interact_Thr.
//
// Return the currently mode if the function succeded (<0 in the other case).
//
int
RemoteGetMode(int Sock_fd)
{
	interact_request_t Request;
	int iResult;

	Request.iType = IOD_MGMT_CMD_GET_NFSP_MODE;
	Request.cTabParam[0] = '\0';
	if (sure_write(Sock_fd, &Request, sizeof (Request)) != sizeof (Request)) {
		Dprintf(L_WARNING, "RemoteSetMode()::sure_write() failed errno=%d", errno);
		return -1;
	}

	if (sure_recv(Sock_fd, &iResult, sizeof (iResult), 0) != sizeof (iResult)) {
		Dprintf(L_WARNING, "RemoteSetMode()::sure_recv() failed errno=%d", errno);
		return -1;
	}

	if (iResult < 0)
		Dprintf(L_WARNING, "There was something wrong during the return mode operation.\n");

	return iResult;
}

//
// Interact_Thr thread : Handles the unix socket (sock_interact): analyse the different request
//  and do operations on iods tab and nfsp kernel
//
// This thread is launched by the Init_IodMgmt function during execution of the first instance of iodmgmt
//
static void *
Interact_Thr(void *dummy __attribute__((unused)))
{
	int Temp_fd;		//socket file descriptor. use fo each connexion return by accept function.
	interact_request_t Request;

	struct sockaddr_un addr;	// Required for the listen and accept.
	int iLen;

	int iResult;		// Result return to the client.

	Dprintf(D_IODMGR, "Interact_thr started");
	//pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	if (listen(sock_interact, 5) < 0) {
		Dprintf(L_WARNING, "Cannot listen on UNIX socket, errno:%d", errno);
		return NULL;
	}

	while (1) {
		iLen = sizeof (addr);	// temp for address
		Temp_fd = accept(sock_interact, (struct sockaddr *) &addr, &iLen);
		if (-1 == Temp_fd) {
			Dprintf(L_WARNING, "accept() failed errno=%d", errno);
			continue;
		}

		while (sure_recv(Temp_fd, ((char *) &Request), sizeof (Request), 0) == sizeof (Request)) {
			switch (Request.iType) {
			case IOD_MGMT_CMD_ADD_IOD:
				iResult = AddIod(Request.cTabParam);
				break;
			case IOD_MGMT_CMD_FREE_IOD:
				iResult = FreeIod(Request.cTabParam);
				break;
			case IOD_MGMT_CMD_LIST_IOD:
				SendList(Temp_fd, &UsedIodList);
				iResult = 1;	//by default
				break;
			case IOD_MGMT_CMD_FLUSH_IOD:
				iResult = Flush();
				break;
			case IOD_MGMT_CMD_STOP_IOD:
				// Quit_IodMgmt() has been registered
				// as an atexit() function so exit()
				// will call it automagically -- pl
				exit(1);
				break;
			case IOD_MGMT_CMD_SET_PING_PARAMETERS:
				iResult = SetConfigParametersByLine(Request.cTabParam);
				break;
			case IOD_MGMT_CMD_GET_PING_PARAMETERS:
				SendConfigParameters(Temp_fd);
				iResult = 1;
				break;
			case IOD_MGMT_CMD_SET_VIODS_NB:
				iResult = SetNbViodByLine(Request.cTabParam);
				break;
			case IOD_MGMT_CMD_SET_NFSP_MODE:
				iResult = SetMode((int) Request.cTabParam[0]);
				break;
			case IOD_MGMT_CMD_GET_NFSP_MODE:
				iResult = iNFSpMode;	//Simply.... ;)
				break;
			default:
				iResult = -1;
			}
			if (sure_write(Temp_fd, &iResult, sizeof (iResult)) != sizeof (iResult)) {
				Dprintf(L_WARNING, "Interact_Thr()::sure_write() failed errno = %d", errno);
				break;
			}
		}		/* while */

		close(Temp_fd);
	}			/* while(1) */
}

		/* **** ********************************************************************************* **** */
		/* **** **************************** Process options and arguments ********************** **** */
		/* **** ********************************************************************************* **** */

//
// Displays usage
//

static void
usage(FILE * f)
{
	fprintf(f,
		"Usage: %s [options] \n"
		"\n"
		" -h             | --help                : this page\n"
		" -F             | --foreground          : don't go background (default: go background)\n"
		" -a \"h:p v\"   | --addiod \"h:p v\"    : add one iod defined by host h:port p and attach to viod v\n"
		" -d \"h:p\"     | --deleteiod \"h:p\"   : unbind and remove one iod defined by host h:port p\n"
		" -i file        | --iodfile file        : add mutliple iod directly from a file\n"
		" -l             | --listconfig          : list all information concerning the iod_mgmt configuration.\n"
		" -f             | --flush               : switch in NFSp DROP mode and flush all iod\n"
/* TODO How can we choose when several iods daemons are present on a same node ? */
		" -s \"n\"       | --setviod \"n\"       : set the viods number, NFSp have to be in DROP mode.\n"
		" -c \"n w p\"   | --configure n w p     : sends exactly n pings, waits w second(s) beetween each\n"
		"                                          and make a pause for p seconds beetween each cycle.\n",
		cPtrProgramName);

	// If we are given stderr, then it is an error so we do not return 0
	exit((f == stderr) ? 1 : 0);
}

//
// Parse the command line from the second instance of iodmgmt
//
static void
parse_command_line(int argc, char *argv[])
{
	static const char *shortopts = "hFa:d:il:c:s:f";
	static const struct option longopts[] = {
		{"help", 0, 0, 'h'},
		{"foreground", 0, 0, 'F'},
		{"addiod", required_argument, 0, 'a'},
		{"deleteiod", required_argument, 0, 'd'},
		{"iodfile", required_argument, 0, 'i'},
		{"listconfig", 0, 0, 'l'},
		{"configure", required_argument, 0, 'c'},
		{"setviod", required_argument, 0, 's'},
		{"flush", 0, 0, 'f'},
		{NULL, 0, 0, 0},
	};
	char c;
	opterr = 0;

	if (argc <= 1) {
		usage(stdout);
	} else {
		// see man 3 getopt.
		while ((c = getopt_long(argc, argv, shortopts, longopts, NULL)) != EOF) {
//                      fprintf(stderr,"mon char is %c\n",c);
			switch (c) {
			case 'h':
				usage(stdout);
				break;
			case 'F':
				foreground = 1;
				break;
			case 'a':	// Add,Drop, Mirror and Bind and iod to an VIOD
				// TODO check how the mirroring step may be implemented
				if (strlen(optarg) > MAXLINELEN) {
					DIE("the line is too long [%s]\n", optarg);
				}
				if (RemoteAddIod(sock_interact, optarg) < 0) {
					Dprintf(L_WARNING, "error while adding an iod\n");
				}
				break;
			case 'd':	// Delete (Unbind an iod from a viod and delete it)
				if (strlen(optarg) > MAXLINELEN) {
					DIE("the line is too long [%s]\n", optarg);
				}
				if (0 > RemoteFreeIod(sock_interact, optarg)) {
					Dprintf(L_WARNING, "error while remote freeing an iod\n");
				}
				break;
			case 'i':	// Configuration by a file
				if (strlen(optarg) > MAXLINELEN) {
					DIE("the line is too long [%s]\n", optarg);
				}
				if (0 > RemoteAddIodsFromFile(sock_interact, optarg)) {
					Dprintf(L_WARNING, "error while adding iods from file\n");
				}
				break;
			case 'f':	// Flush everything
				if (0 > RemoteFlush(sock_interact)) {
					Dprintf(L_WARNING, "error while remote-flushing");
				}
				break;
			case 'l':	//List informations about iod and viod
				if (0 > RemoteDisplayIodsList(sock_interact)) {
					Dprintf(L_WARNING, "error while displaying remote iod list");
				}
				if (0 > RemoteDisplayConfigParameters(sock_interact)) {
					Dprintf(L_WARNING, "error while displaying configuration parameters\n");
				}
				break;
			case 'c':	//Configure the Ping thread parameters.
				if (strlen(optarg) > MAXLINELEN) {
					DIE("the line is too long [%s]\n", optarg);
				}
				if (0 > RemoteSetConfigParameters(sock_interact, optarg)) {
					Dprintf(L_WARNING, "error while initializing the ping parameters\n");
				}
				break;
			case 's':	//Configure the viods number.
				if (strlen(optarg) > MAXLINELEN) {
					DIE("the line is too long [%s]\n", optarg);
				}
				if (0 > RemoteSetNbViod(sock_interact, optarg)) {
					Dprintf(L_WARNING, "error while initializing the viod\n");
				}
				break;
			default:
				usage(stderr);
			}
		}		// End of switch and while opts.
	}
}

		/* **** ******************************************************************************** **** */
		/* **** ********************************* Miscellaneous ******************************** **** */
		/* **** ******************************************************************************** **** */

//
// My xstrdup
// Equal to xstrdup in xstrdup.c but not require to link libnfs.a with iod_mgmt
// Allocates a char* and test it before to apply the strcpy function.
//
static char *
xstrdup(const char *string)
{
	char *cPtrTemp;

	cPtrTemp = malloc(strlen(string) + 1);
	if (cPtrTemp) {
		strcpy(cPtrTemp, string);
	} else {
		Dprintf(L_WARNING, "Error in xstrdup(): cannot allocate a char buffer Program:%s", cPtrProgramName);
	}

	return cPtrTemp;
}

static void
Quit_SignalHandler(int n)
{
	fprintf(stderr, "Received signal %d\n", n);
	exit(1);
}

//
// Create the Ping thread and the Interact thread
// Init the different socket to iods and to iod_mgmt Interact Thread
//
static void
Init_IodMgmt(void)
{
	Dprintf(D_IODMGR, "Beginning of Init Iod Mgmt");

	// To properly quit
	atexit(Quit_IodMgmt);
	signal(SIGHUP, Quit_SignalHandler);	// call exit(1)
	signal(SIGINT, Quit_SignalHandler);	// call exit(1)
	signal(SIGTERM, Quit_SignalHandler);	// call exit(1)

	// Initialize the UDP socket: dies if it can't be opened
	iods_udp_fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (iods_udp_fd < 0) {
		Dprintf(L_FATAL, "cannot open UDP socket errno=%d", errno);
		exit(1);
	}

#if IOD_MGMT_WITH_NFSP
	// Initialize NFSp Kernel
	if (nfspd_initialize() < 0) {
		Dprintf(L_WARNING, "Something is wrong with nfspd_initialize() errno=%d", errno);
	}
#endif

	if (SetMode(IOD_MGMT_MODE_DROP)) {
		Dprintf(L_WARNING, "cannot set NFSp in DROP mode");
	}

	// Create UNIX socket to discuss between the different instance of iod_mgmt
	do {
		struct sockaddr_un sun;
		char *home;

		sock_interact = socket(PF_UNIX, SOCK_STREAM, 0);
		if (sock_interact < 0) {
			Dprintf(L_FATAL, "cannot open UNIX socket errno=%d", errno);
			exit(1);
		}

		sun.sun_family = AF_UNIX;
		home = getenv(SOCK_INTERACT_ENV_VAR);
		if (!home) {
			Dprintf(L_FATAL, "%s must be set", SOCK_INTERACT_ENV_VAR);
			exit(1);
		}
		if (sizeof(sun.sun_path) < strlen(home) + 1 + strlen(SOCK_INTERACT_NAME) + 1) {
			Dprintf(L_FATAL, "path to UNIX control socket too long (>%d)", sizeof(sun.sun_path));
			exit(1);
		}
		strcpy(sun.sun_path, home);
		strcat(sun.sun_path, "/");
		strcat(sun.sun_path, SOCK_INTERACT_NAME);

		if (-1 == bind(sock_interact, (struct sockaddr *) &sun, sizeof(sun))) {
			Dprintf(L_FATAL, "Init_IodMgmt() cannot bind() UNIX socket errno=%d", errno);
			exit(1);
		}
	} while (0);

	do {
		// Initialize iods tabs and store in unused list.
		int i;
		INIT_LIST_HEAD(&FreeIodList);
		for (i = 0; i < MAX_IODS; i++) {
			iods[i].check_ok = 0;
			iods[i].iRelated_viod = -1;
			iods[i].iKernelHandle = -1;
			list_add_tail(&(iods[i].FromAndToList), &FreeIodList);
		}
		INIT_LIST_HEAD(&UsedIodList);
	}
	while (0);

	foreground = 1;
	if (!foreground) {
		// user wants a daemon
		pid_t pid = fork();
		switch (pid) {
		case -1:	// error

			Dprintf(L_FATAL, "cannot daemonize (fork failed) errno=%d", errno);
			exit(1);
			break;
		case 0:	// fork()'ed process
			setsid();
			close(0);
			close(1);
			close(2);
			break;
		default:	// original process
			exit(0);
		}
	}

	do {
		int ret;
		pthread_t thr;

		ret = pthread_create(&thr, NULL, Ping_Thr, NULL);
		if (ret) {
			Dprintf(L_FATAL, "cannot spawn a thread");
			exit(1);
		}
		ret = pthread_detach(thr);
		if (ret) {
			Dprintf(L_FATAL, "cannot detach a thread");
			exit(1);
		}
		Dprintf(D_IODMGR, "Ping thread %ld up and running", thr);
	} while (0);

	Interact_Thr(NULL);
//      Dprintf(D_IODMGR, "Interact thread %ld up and running", thr);
}

static void
Quit_IodMgmt(void)
{
	// TODO : Synchronize cancel request (be carreful not to cancel a pthread that is holding a mutex lock)
	//pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	//pthread_cancel(Ping_Thr);
	//pthread_cancel(Interact_Thr)
	//

#if IOD_MGMT_WITH_NFSP
	// Mode USER
	// Terminate the soket connection with nfspd
	if (nfspd_terminate() < 0) {
		Dprintf(L_WARNING, "Unable to close nfspd control socket,(errno = %d)", errno);
	}
#endif
	// Close UDP socket.
	if (close(iods_udp_fd)) {
		Dprintf(L_WARNING, "unable to close UDP socket errno=%d\n", errno);
	}
	// Close UNIX socket.
	if (close(sock_interact)) {
		Dprintf(L_WARNING, "unable to close UNIX socket errno=%d\n", errno);
	}
	// Unlink the related file from the UNIX socket.
	{
		char sunpath[256];
		char *home;

		home = getenv(SOCK_INTERACT_ENV_VAR);
		if (!home) {
			Dprintf(L_FATAL, "%s must be set", SOCK_INTERACT_ENV_VAR);
			exit(1);
		}
		if (sizeof(sunpath) < strlen(home) + 1 + strlen(SOCK_INTERACT_NAME) + 1) {
			Dprintf(L_FATAL, "path to UNIX control socket too long (>%d)", sizeof(sunpath));
			exit(1);
		}
		strcpy(sunpath, home);
		strcat(sunpath, "/");
		strcat(sunpath, SOCK_INTERACT_NAME);

		if (unlink(sunpath)) {
			Dprintf(L_WARNING, "unable to unlink '%s' errno=%d\n", sunpath, errno);
		}
	}
}

/* **** ****************************************************************************************************************** **** */
/* **** ********************************************** Main ************************************************************** **** */
/* **** ****************************************************************************************************************** **** */

int
main(int argc, char **argv)
{

	Dprintf(D_IODMGR, "Compilation : " __DATE__ " at " __TIME__ " with " __VERSION__ "");

	cPtrProgramName = xstrdup(argv[0]);
	if ((argc == 2) && (!strcmp(argv[1], "start"))) {
		// have to test if the daemon already run ????
		Init_IodMgmt();	// set up Ping_thr, Interact_thr and open the different socket

	} else {
		// Create the UNIX socket used to make the different instances of iod_mgmt
		do {
			struct sockaddr_un sun;
			char *home;

			sock_interact = socket(PF_UNIX, SOCK_STREAM, 0);
			if (sock_interact < 0) {
				Dprintf(L_FATAL, "cannot open UNIX socket errno=%d", errno);
				exit(1);
			}

			sun.sun_family = AF_UNIX;
			home = getenv(SOCK_INTERACT_ENV_VAR);
			if (!home) {
				Dprintf(L_FATAL, "%s must be set", SOCK_INTERACT_ENV_VAR);
				exit(1);
			}
			if (sizeof(sun.sun_path) < strlen(home) + 1 + strlen(SOCK_INTERACT_NAME) + 1) {
				Dprintf(L_FATAL, "path to UNIX control socket too long (>%d)", sizeof(sun.sun_path));
				exit(1);
			}
			strcpy(sun.sun_path, home);
			strcat(sun.sun_path, "/");
			strcat(sun.sun_path, SOCK_INTERACT_NAME);

			if (-1 == connect(sock_interact, (struct sockaddr *) &sun, sizeof (sun))) {
				if (errno == ECONNREFUSED) {
					// If the connection was refused, then
					// it means that the Interact thread does not run.
					if (close(sock_interact)) {
						Dprintf(L_WARNING, "cannot close UNIX socket\n");
					}
					fprintf(stderr, "errno=%d, iod_mgmt did not start.\n" "Please use 'iod_mgmt start' command\n", errno);
				} else {
					Dprintf(L_FATAL, "Cannot connect to the socket between Interact Thread" " and iod_mgmt process errno=%d\n", errno);
					exit(1);
				}

			}
		} while (0);

		if (argc == 2) {
			if (!strcmp(argv[1], "stop")) {
				interact_request_t Request;

				Request.iType = IOD_MGMT_CMD_STOP_IOD;
				Request.cTabParam[0] = '\0';
				if (sure_write(sock_interact, &Request, sizeof (Request)) != sizeof (Request)) {
					Dprintf(L_WARNING, "something nasty happened in sure_write in function main (errno = %d).\n", errno);
					return -1;
				}
			} else if (!strcmp(argv[1], "drop")) {
				RemoteSetMode(sock_interact, IOD_MGMT_MODE_DROP);
			} else if (!strcmp(argv[1], "forward")) {
				RemoteSetMode(sock_interact, IOD_MGMT_MODE_FORWARD);
			} else {
				parse_command_line(argc, argv);
			}
		} else {
			parse_command_line(argc, argv);
		}

		if (close(sock_interact)) {
			Dprintf(L_WARNING, "unable to close UNIX socket");
		}
	}
	return 0;
}
